<?php

	/**
	 * Modul zur Zahlungsabwicklung über die PayPal REST API
	 * @author Daschmi (daniel@maennchen1.de)
	 */
	class wpsg_mod_paypalapi extends wpsg_mod_basic
	{

		var $apiContext = false;
		
		var $lizenz = 1;
		var $id = 10;		
		var $hilfeURL = 'http://wpshopgermany.de/?p=1728009';
		var $version = "4.4.7";
		var $free = false;
		
		private $currency = "EUR";
		
		// Wenn true, dann keine direkte Fehlerbehandlung
		private $api_try = false;
		
		/**
		 * Costructor 
		 */
		public function __construct()
		{
				
			parent::__construct();
				
			$this->name = __('PayPal PLUS', 'wpsg');
			$this->group = __('Zahlungsarten', 'wpsg');
			$this->desc = __('Ermöglicht die Zahlungsart PayPal und PayPal PLUS, über die REST API Schnittstelle und Login mit PayPal Express.', 'wpsg');
			
		} // public function __construct()
				
		public function init() {

            spl_autoload_register(array($this, 'spl_autoload'));
		
			// Wenn Währungscode in den Einstellungen gesetzt ist dann diesen nehmen, sonst bleibt alles wie bisher bei EUR
			if (wpsg_isSizedString($this->shop->get_option('wpsg_mod_paypalapi_currency'))) $this->currency = $this->shop->get_option('wpsg_mod_paypalapi_currency');
			
		} // public function init()
				
		public function install() {

            $this->shop->checkDefault('wpsg_mod_paypalapi_version', 'v2');
			$this->shop->checkDefault('wpsg_mod_paypalapi_aktiv', '1');
			$this->shop->checkDefault('wpsg_mod_paypalapi_paypalexpress', '1');
			$this->shop->checkDefault('wpsg_mod_paypalapi_plus_hint', __('Zahlen Sie die Bestellung mittels ihres PayPal Kontos', 'wpsg'), false, true);
			$this->shop->checkDefault('wpsg_mod_paypalapi_paypalexpress_name', __('PayPal', 'wpsg'), false, true);
			$this->shop->checkDefault('wpsg_mod_paypalapi_paypalexpress_subject', 'O%order_id% - K%kunde_id%', false, true);
			$this->shop->checkDefault('wpsg_mod_paypalapi_paypalexpress_details', '1');
			$this->shop->checkDefault('wpsg_mod_paypalapi_currency', 'EUR');
					 			
			$this->shop->checkDefault('wpsg_mod_paypalapi_experience_label', '%shopinfo_name%', false, true);
			$this->shop->checkDefault('wpsg_mod_paypalapi_experience_locale_code', 'DE');
			
			$this->shop->checkDefault('wpsg_mod_paypalapi_plus_aktiv', '1');		
									
			$this->shop->checkDefault('wpsg_mod_paypalapi_stornostate', array('400' => '1', '500' => '1'));
							
		} // public function install()
		
		public function settings_edit() {

			$pages = get_pages();
				
			$arPages = array(
					'-1' => __('Neu anlegen und zuordnen', 'wpsg')
			);
				
			foreach ($pages as $k => $v)
			{
				$arPages[$v->ID] = $v->post_title.' (ID:'.$v->ID.')';
			}
				
			$this->shop->view['pages'] = $arPages;
					
			$this->shop->render(WPSG_PATH_VIEW.'/mods/mod_paypalapi/settings_edit.phtml');
				
		} // public function settings_edit()
		
		public function settings_save() {
			
			foreach ($_REQUEST['wpsg_mod_paypalapi_stornostate'] as $k => $v) {
				
				$bKeyValid = wpsg_checkInput($k, WPSG_SANITIZE_INT);
				$bValueValid = wpsg_checkInput($v, WPSG_SANITIZE_CHECKBOX);
				
				if (!$bKeyValid || !$bValueValid) {
					
					$this->shop->addBackendError(__('Ihre Eingaben in den markierten Feldern waren ungültig, bitte überprüfen.', 'wpsg'));
					
					$_SESSION['sanitization_err_fields']['wpsg_mod_paypalapi_stornostate'] = 0;
					
				} else {
					
					$this->shop->update_option('wpsg_mod_paypalapi_stornostate', $_REQUEST['wpsg_mod_paypalapi_stornostate']); // Sanitized at begin of function call
					
				}
				
			}

            $this->shop->update_option('wpsg_mod_paypalapi_version', $_REQUEST['wpsg_mod_paypalapi_version'], false, false, WPSG_SANITIZE_VALUES, ['v1', 'v2']);
			$this->shop->update_option('wpsg_mod_paypalapi_clientid', $_REQUEST['wpsg_mod_paypalapi_clientid'], false, false, WPSG_SANITIZE_APIKEY);
			$this->shop->update_option('wpsg_mod_paypalapi_secret', $_REQUEST['wpsg_mod_paypalapi_secret'], false, false, WPSG_SANITIZE_APIKEY);
			$this->shop->update_option('wpsg_mod_paypalapi_sandbox', $_REQUEST['wpsg_mod_paypalapi_sandbox'], false, false, WPSG_SANITIZE_CHECKBOX);
			$this->shop->update_option('wpsg_mod_paypalapi_sandbox_clientid', $_REQUEST['wpsg_mod_paypalapi_sandbox_clientid'], false, false, WPSG_SANITIZE_APIKEY);
			$this->shop->update_option('wpsg_mod_paypalapi_sandbox_secret', $_REQUEST['wpsg_mod_paypalapi_sandbox_secret'], false, false, WPSG_SANITIZE_APIKEY);
			$this->shop->update_option('wpsg_mod_paypalapi_currency', $_REQUEST['wpsg_mod_paypalapi_currency'], false, false, WPSG_SANITIZE_TEXTFIELD);
			$this->shop->update_option('wpsg_mod_paypalapi_webhook_log', $_REQUEST['wpsg_mod_paypalapi_webhook_log'], false, false, WPSG_SANITIZE_CHECKBOX);
			$this->shop->update_option('wpsg_mod_paypalapi_webhook_logfile', $_REQUEST['wpsg_mod_paypalapi_webhook_logfile'], false, false, WPSG_SANITIZE_PATH);
			
			$this->shop->update_option('wpsg_mod_paypalapi_gebuehr', $_REQUEST['wpsg_mod_paypalapi_gebuehr'], false, false, WPSG_SANITIZE_FLOAT);
			$this->shop->update_option('wpsg_mod_paypalapi_mwst', $_REQUEST['wpsg_mod_paypalapi_mwst'], false, false, WPSG_SANITIZE_TAXKEY);
			$this->shop->update_option('wpsg_mod_paypalapi_mwstland', $_REQUEST['wpsg_mod_paypalapi_mwstland'], false, false, WPSG_SANITIZE_CHECKBOX);
						
			$this->shop->update_option('wpsg_mod_paypalapi_aktiv', $_REQUEST['wpsg_mod_paypalapi_aktiv'], false, false, WPSG_SANITIZE_CHECKBOX);
			$this->shop->update_option('wpsg_mod_paypalapi_paypalexpress', $_REQUEST['wpsg_mod_paypalapi_paypalexpress'], false, false, WPSG_SANITIZE_CHECKBOX);
			$this->shop->update_option('wpsg_mod_paypalapi_paypalexpress_forceSSL', $_REQUEST['wpsg_mod_paypalapi_paypalexpress_forceSSL'], false, false, WPSG_SANITIZE_CHECKBOX);
			
			$this->shop->update_option('wpsg_mod_paypalapi_paypalexpress_name', $_REQUEST['wpsg_mod_paypalapi_paypalexpress_name'], false, true, WPSG_SANITIZE_TEXTFIELD);
			$this->shop->update_option('wpsg_mod_paypalapi_plus_hint', $_REQUEST['wpsg_mod_paypalapi_plus_hint'], false, true, WPSG_SANITIZE_TEXTAREA);
			$this->shop->update_option('wpsg_mod_paypalapi_paypalexpress_subject', $_REQUEST['wpsg_mod_paypalapi_paypalexpress_subject'], false, false, WPSG_SANITIZE_TEXTFIELD);
			
			$this->shop->update_option('wpsg_mod_paypalapi_autostart', $_REQUEST['wpsg_mod_paypalapi_autostart'], false, false, WPSG_SANITIZE_CHECKBOX);
			$this->shop->update_option('wpsg_mod_paypalapi_paypalexpress_details', $_REQUEST['wpsg_mod_paypalapi_paypalexpress_details'], false, false, WPSG_SANITIZE_CHECKBOX);
			
			$this->shop->update_option('wpsg_mod_paypalapi_experience_label', $_REQUEST['wpsg_mod_paypalapi_experience_label'], false, true, WPSG_SANITIZE_TEXTFIELD);
			$this->shop->update_option('wpsg_mod_paypalapi_experience_logo', $_REQUEST['wpsg_mod_paypalapi_experience_logo'], false, false, WPSG_SANITIZE_URL);
			$this->shop->update_option('wpsg_mod_paypalapi_experience_locale_code', $_REQUEST['wpsg_mod_paypalapi_experience_locale_code'], false, false, WPSG_SANITIZE_VALUES, ['DE', 'GB', 'FR']);

			$this->shop->update_option('wpsg_mod_paypalapi_plus_aktiv', $_REQUEST['wpsg_mod_paypalapi_plus_aktiv'], false, false, WPSG_SANITIZE_CHECKBOX);
			$this->shop->update_option('wpsg_mod_paypalapi_plus_basket', $_REQUEST['wpsg_mod_paypalapi_plus_basket'], false, false, WPSG_SANITIZE_CHECKBOX);
			
			// WebHook
			if (wpsg_isSizedInt($_REQUEST['wpsg_mod_paypalapi_create'])) $this->createWebHook();

			// PaymentExperience
			if (wpsg_isSizedString($_REQUEST['wpsg_mod_paypalapi_clientid']) && wpsg_isSizedString($_REQUEST['wpsg_mod_paypalapi_secret'])) {
				
				$json_data = $this->api_getWebExperience();
				 
				$paymentExperienceID = false;
				
				foreach ($json_data as $p) if ($p['name'] == 'wpShopGermany') $paymentExperienceID = $p['id'];
				
				if ($paymentExperienceID === false) {
					
					// PaymentExperience existiert nicht => anlegen
					$json_data = $this->api_createWebExperience();
				
					$this->shop->update_option('wpsg_mod_paypalapi_paymentExperience', $json_data['id'], false, WPSG_SANITIZE_TEXTFIELD);
					
				} else
				{
				
					$this->shop->update_option('wpsg_mod_paypalapi_paymentExperience', $paymentExperienceID);
					
					$this->api_updateWebExperience($paymentExperienceID);
										
				}
				
			}
			
		} // public function settings_save()
		
		public function systemcheck(&$arData) {
			
			if ($this->shop->get_option('wpsg_mod_paypalapi_aktiv') === '1') {
				
				$arPflicht = $this->shop->loadPflichtFeldDaten();
				
				if (!isset($arPflicht['plz']) || $arPflicht['plz'] !== '0') {
					
					$arData[] = array(
						'wpsg_mod_paypal_zip',
						wpsg_ShopController::CHECK_WARNING,
						wpsg_translate(
							__('Bei verwendung von PayPal als Zahlungsart sollte die PLZ ein Pflichtfeld sein, damit die API funktioniert. <a href="#1#">hier</a> konfigurieren.', 'wpsg'),
							WPSG_URL_WP.'wp-admin/admin.php?page=wpsg-Admin&subaction=kundendaten'
						)
					);
					
				}

				$nProductWithoutAnr = intval($this->db->fetchOne("SELECT COUNT(*) FROM `".WPSG_TBL_PRODUCTS."` WHERE `anr` = '' AND `deleted` != '1' AND `lang_parent` = '0' "));
				
				if ($nProductWithoutAnr > 0 && $this->shop->get_option('wpsg_mod_paypalapi_paypalexpress_details')) {
					
					$arData[] = array(
						'wpsg_mod_paypal_apidata',
						wpsg_ShopController::CHECK_ERROR,
						wpsg_translate(
							__('Es wurden #1# Produkt(e) ohne Artikelnummer gefunden. PayPal PLUS erfordert mit der Option "Warenkorbdetails übertragen" Artikelnummern.', 'wpsg'),
							$nProductWithoutAnr
						)
					);
					
				}
				
				if ($this->shop->get_option('wpsg_mod_paypalapi_sandbox') === '1') {
					
					if (!wpsg_isSizedString($this->shop->get_option('wpsg_mod_paypalapi_sandbox_clientid')) && !wpsg_isSizedString($this->shop->get_option('wpsg_mod_paypalapi_sandbox_secret'))) {
						
						$arData[] = array(
							'wpsg_mod_paypal_apidata',
							wpsg_ShopController::CHECK_ERROR,
							wpsg_translate(
								__('Das Modul "PayPal Plus" ist aktiviert und es wurden keine API Zugangsdaten hinterlegt (Sandbox). Gehen Sie in die <a href="#1#">Einstellungen</a> des Moduls um die API Daten zu hinterlegen.', 'wpsg'),
								WPSG_URL_WP.'wp-admin/admin.php?page=wpsg-Admin&action=module&modul=wpsg_mod_paypalapi'
							)
						);
						
					}
					
				} else {
				
					if (!wpsg_isSizedString($this->shop->get_option('wpsg_mod_paypalapi_clientid')) && !wpsg_isSizedString($this->shop->get_option('wpsg_mod_paypalapi_secret'))) {
						
						$arData[] = array(
							'wpsg_mod_paypal_apidata',
							wpsg_ShopController::CHECK_ERROR,
							wpsg_translate(
								__('Das Modul "PayPal Plus" ist aktiviert und es wurden keine API Zugangsdaten hinterlegt. Gehen Sie in die <a href="#1#">Einstellungen</a> des Moduls um die API Daten zu hinterlegen.', 'wpsg'),
								WPSG_URL_WP.'wp-admin/admin.php?page=wpsg-Admin&action=module&modul=wpsg_mod_paypalapi'
							)
						);
						
					}
					
				}
				
				if ($this->shop->get_option('wpsg_mod_paypalapi_currency') !== 'EUR') {
					
					$arData[] = array(
						'wpsg_mod_paypal_currency',
						wpsg_ShopController::CHECK_ERROR,
						wpsg_translate(
							__('Der Währungscode sollte bei der Verwendung mit PayPal auf "EUR" stehen. Gehen Sie in die <a href="#1#">Einstellungen</a> des Moduls um dies zu korrigieren.', 'wpsg'),
							WPSG_URL_WP.'wp-admin/admin.php?page=wpsg-Admin&action=module&modul=wpsg_mod_paypalapi'
						)
					);
					
				}
				
			}
			
		}
		
        public function wpsg_deinstall_sites() {

            wp_delete_post($this->shop->get_option('wpsg_page_mod_paypalapi_success'));
            wp_delete_post($this->shop->get_option('wpsg_page_mod_paypalapi_error'));

        } // public function wpsg_deinstall_sites()

		public function setOrderStatus($order_id, $status_id, $inform) {
		
			$oOrder = wpsg_order::getInstance($order_id);
			
			if ($oOrder->getPaymentID() == $this->id) {
				
				$arStornoState = (array)$this->shop->get_option('wpsg_mod_paypalapi_stornostate');
			
				if (array_key_exists($status_id, $arStornoState) && $arStornoState[$status_id] == '1') {
					
					if (wpsg_isSizedString($oOrder->getMeta('wpsg_mod_paypalapi_saleid'))) {

                        $state = $this->getState($order_id);

                        if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v2') {

                            if ($state[1] === 'COMPLETED') {

                                $bStorno = $this->stornoOrder($order_id);

                                if ($bStorno === true) $this->shop->addBackendMessage(__('Zahlung erfolgreich über die PayPal API storniert.', 'wpsg'));
                                else $this->shop->addBackendError(__('Es gab ein Problem bei der Stornierung der Zahlung über die PayPal API. Bitte Bestellprotokoll beachten.', 'wpsg'));

                            } else $this->shop->addBackendError(__('PayPal Rückzahlung nicht möglich.'));

                        } else {

                            $bStorno = $this->stornoOrder($order_id);

                            if ($bStorno === true) $this->shop->addBackendMessage(__('Zahlung erfolgreich über die PayPal API storniert.', 'wpsg'));
                            else $this->shop->addBackendError(__('Es gab ein Problem bei der Stornierung der Zahlung über die PayPal API. Bitte Bestellprotokoll beachten.', 'wpsg'));

                        }
						
					}
					
				}	

			}
			 				
		}
				
		public function addPayment(&$arPayment) {
 					
			if (is_admin() || $this->shop->get_option('wpsg_mod_paypalapi_aktiv') == '1') { 

				$arPayment[$this->id] = array(
					'id' => $this->id,
					'name' => __($this->shop->get_option('wpsg_mod_paypalapi_paypalexpress_name'), 'wpsg'),
					'price' => $this->shop->get_option('wpsg_mod_paypalapi_gebuehr'),
					'tax_key' => $this->shop->get_option('wpsg_mod_paypalapi_mwst'),
					'mwst_null' => $this->shop->get_option('wpsg_mod_paypalapi_mwstland'),
					'hint' => __($this->shop->get_option('wpsg_mod_paypalapi_plus_hint'), 'wpsg'),
					'logo' => $this->shop->getRessourceURL('mods/mod_paypalapi/gfx/logo_100x25.png')
				);
			 			
			}
			
		} // public function addPayment(&$arPayment)
				 
		public function checkout_handlePayment() {
		
			// Wenn 0,- dann nix machen
			$arBasket = $this->shop->basket->toArray();
			if ($arBasket['arCalculation']['sum']['topay_brutto'] <= 0) return false;
			
			// Wenn PayPal nicht drin ist, dann nichts machen
			if (!array_key_exists($this->id, $this->shop->arPayment)) return false;
			
			$this->shop->view['wpsg_mod_paypalapi']['show'] = 0;
							
			// Wenn eine andere Zahlungsart als PayPal gewählt wurde und Zahlungskosten entstanden sind,
			// dann lösche ich die Zahlung und lasse neu wählen			
			if ($this->id != $_SESSION['wpsg']['checkout']['payment'] && isset($this->shop->view['basket']['sum']['preis_payment'])  && $this->shop->view['basket']['sum']['preis_payment'] != 0) {
			
				unset($_SESSION['wpsg']['checkout']['payment']);
				
				$this->shop->basket->initFromSession(true);
				$this->shop->view['basket'] = $this->shop->basket->toArray();
								
			}
						
			// Zahlung über kompletten Betrag schon vorbereitet
			if (isset($_SESSION['wpsg']['checkout']['payment_amount']) && $_SESSION['wpsg']['checkout']['payment'] == $this->id) {
				
				if (wpsg_tf($_SESSION['wpsg']['checkout']['payment_amount']) == $this->shop->view['basket']['sum']['preis_gesamt_brutto']) {
					
					if ($this->shop->hasMod('wpsg_mod_willcollect')) {
							
						$arCheckout = array(
							'payment' => $_SESSION['wpsg']['checkout']['payment'],
							'shipping' => $_SESSION['wpsg']['checkout']['shipping']
						);
					
						$error = false; $step = 2;
						$this->shop->callMod('wpsg_mod_willcollect', 'checkCheckout', array(&$step, &$error, &$arCheckout));
					
						if ($error === true) {
							
							$this->shop->view['wpsg_mod_paypalapi']['error'] = __('Die Zahlungsart PayPal ist leider nicht mit der Versandart "Selbstabholung" möglich. Ihre Zahlung wurde storniert und ihr Konto nicht belastet. Bitte wählen Sie eine andere Zahlungsart aus.', 'wpsg');
							
							$this->setViewForPayPalPLus();
							$this->shop->view['wpsg_mod_paypalapi']['show'] = 3;
								
						} else {

							$this->shop->view['wpsg_mod_paypalapi']['show'] = 1;

						}
					
					} else {
					
						$this->shop->view['wpsg_mod_paypalapi']['show'] = 1;
						
					}
					
				} else {
			
					if (wpsg_isSizedInt($this->shop->get_option('wpsg_mod_paypalapi_plus_aktiv'))) {
						 
						$this->setViewForPayPalPLus();						
						$this->shop->view['wpsg_mod_paypalapi']['show'] = 3;
						
					} else {
					
						$this->shop->view['wpsg_mod_paypalapi']['show'] = 2;
						$this->shop->render(WPSG_PATH_VIEW.'/mods/mod_paypalapi/checkout_handlePayment.phtml');
						
						return false;
						
					}
					 
				}
				
			} else {
				
				if (wpsg_isSizedInt($this->shop->get_option('wpsg_mod_paypalapi_plus_aktiv'))) {
				
					$this->setViewForPayPalPLus();
					$this->shop->view['wpsg_mod_paypalapi']['show'] = 3;
						
				} else {
					
					return false;
					
				}
				
			}
			
			if (isset($_SESSION['wpsg']['checkout']['payment']) && $_SESSION['wpsg']['checkout']['payment'] != '10') { 
			
				$this->shop->view['wpsg_mod_paypalapi']['error'] = __('Sie hatten eventuell bereits eine Zahlungsart gewählt. Aufgrund technischer Einschränkungen kann es sein, dass diese Auswahl hier erneut durchgeführt werden muss.', 'wpsg');
				
			}
			 
			$this->shop->render(WPSG_PATH_VIEW.'/mods/mod_paypalapi/checkout_handlePayment.phtml');
			
			return -2;
							 			
		} // public function checkout_handlePayment() 
		
		public function mail_payment()
		{
		
			if ($this->shop->view['basket']['checkout']['payment'] != $this->id) return;
				
			if ($this->shop->htmlMail === true)
			{
		
				echo '<a href="'.$this->shop->getDoneURL($this->shop->view['o_id']).'">'.__('Zahlungslink', 'wpsg').'</a>'.__(', um die Zahlung durchzuführen', 'wpsg');
		
			}
			else
			{
					
				echo wpsg_pad_right(__('Zahlungslink', 'wpsg').':', 35).$this->shop->getDoneURL($this->shop->view['o_id'])." 
";				
		
			}
				
		} // public function mail_payment()
		
		public function order_ajax() {

			if (!wpsg_checkInput($_REQUEST['edit_id'], WPSG_SANITIZE_INT)) throw \wpsg\Exception::getSanitizeException(); else $order_id = $_REQUEST['edit_id'];
 			if (!wpsg_checkInput($_REQUEST['do'], WPSG_SANITIZE_VALUES, ['refund', 'refresh'])) throw \wpsg\Exception::getSanitizeException(); else $do = $_REQUEST['do'];

            $oOrder = wpsg_order::getInstance(intval($order_id));

			if ($do === 'refund') {
			
				$bStorno = $this->stornoOrder($order_id);
				
				if ($bStorno === true) $this->shop->addBackendMessage(__('Zahlung erfolgreich über die PayPal API storniert.', 'wpsg'));
				else $this->shop->addBackendError(__('Es gab ein Problem bei der Stornierung der Zahlung über die PayPal API. Bitte Bestellprotokoll beachten.', 'wpsg'));
				
				$this->shop->redirect(
					wpsg_admin_url('Order', 'view', ['edit_id' => $order_id])
				);
			
			} else if ($do === 'refresh') {

			    try {

			        $state = $this->getState($order_id);

                } catch (\Exception $e) { }
 
				if (is_array($state)) {

                    list($payment_state, $sale_state, $result) = $state;

                    $oOrder->addLogEntry('PayPal Status', print_r($result, true));

                    $this->shop->addBackendMessage(wpsg_translate(__('Status der Zahlung (#1#) erfolgreich abgefragt.', 'wpsg'), $payment_state.'/'.$sale_state));

                } else {

                    $this->shop->addBackendError(__('Status der Zahlung konnte nicht abgefragt werden.', 'wpsg'));
					
                }
				
				$this->shop->redirect(
					wpsg_admin_url('Order', 'view', ['edit_id' => $order_id])
				);
				
			}
			
		} // public function order_ajax()
		
		public function order_done(&$order_id, &$done_view) {

			$oOrder = wpsg_order::getInstance($order_id);

			// Bestellungen mit 0 geben nix aus
			if ($oOrder->getToPay(WPSG_BRUTTO) <= 0) return;

			// Bestellungen die nicht mit PayPal API beglichen werden sollen ignorieren
			if ($oOrder->getPaymentID() != $this->id) return;

			$wpsg_mod_paypalapi_paymentid = $oOrder->getMeta('wpsg_mod_paypalapi_paymentid');

			if (wpsg_isSizedString($wpsg_mod_paypalapi_paymentid)) {

                if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v1') {

                    $payment_info = $this->api_getPaymentInfo($wpsg_mod_paypalapi_paymentid, true);
                    $amount_approved = floatval($payment_info['transactions'][0]['amount']['total']);

                } else {

                    try {

                        $payment_info = $this->api_getPaymentInfo_v2($wpsg_mod_paypalapi_paymentid, true);
                        
                        $amount_approved = floatval($payment_info['purchase_units'][0]['amount']['value']);

                    } catch (\Exception $e) {

                    }

                }

				//if ($amount_approved === floatval($oOrder->getAmount()) && !wpsg_isSizedString($oOrder->getMeta('wpsg_mod_paypalapi_saleid')))
				// Ausführen bei Approved und Created! Wenn der Iframe angezeigt wird, wird die alte PayMent ID verworfen!
				
				$wpsg_mod_paypalapi_saleid = $oOrder->getMeta('wpsg_mod_paypalapi_saleid');

				if ($amount_approved === floatval($oOrder->getAmount()) && !wpsg_isSizedString($wpsg_mod_paypalapi_saleid)) {

				    try {

				        $payment_info_execute = $this->api_executePayment_v2($oOrder->getMeta('wpsg_mod_paypalapi_paymentid'), $oOrder->getMeta('wpsg_mod_paypalapi_payer_id'));

                        if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v1') {

                            if (wpsg_isSizedString($payment_info_execute['transactions'][0]['related_resources'][0]['sale']['id']))
                            {

                                $sale_id = $payment_info_execute['transactions'][0]['related_resources'][0]['sale']['id'];

                                $oOrder->setMeta('wpsg_mod_paypalapi_saleid', $sale_id);
                                $oOrder->setMeta('wpsg_mod_paypalapi_salestate', $payment_info_execute['transactions'][0]['related_resources'][0]['sale']['state']);
                                $oOrder->setMeta('wpsg_mod_paypalapi_paymentstate', $payment_info_execute['state']);

                                if ($payment_info_execute['state'] == 'approved')
                                {

                                    $this->shop->view['wpsg_mod_paypalapi']['done'] = '2';
                                    $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_paypalapi/order_done.phtml');

                                    return;

                                }

                                else $payment_info = $payment_info_execute;

                            }

                        } else {

                            if (wpsg_isSizedString($payment_info_execute['purchase_units'][0]['payments']['captures'][0]['id'])) {

                                $sale_id = $payment_info_execute['purchase_units'][0]['payments']['captures'][0]['id'];

                                $oOrder->setMeta('wpsg_mod_paypalapi_saleid', $sale_id);
                                $oOrder->setMeta('wpsg_mod_paypalapi_salestate', $payment_info_execute['purchase_units'][0]['payments']['captures'][0]['status']);
                                $oOrder->setMeta('wpsg_mod_paypalapi_paymentstate', $payment_info_execute['status']);

                                if ($payment_info_execute['status'] === 'COMPLETED') {

                                    $this->shop->view['wpsg_mod_paypalapi']['done'] = '2';
                                    $this->shop->render(WPSG_PATH_VIEW.'/mods/mod_paypalapi/order_done.phtml');

                                    return;

                                }
                                else $payment_info = $payment_info_execute;

                            }

                        }

                    } catch (\Exception $e) {

				        wpsg_debug($e->getMessage());
				        // ...

                    }
					
				}
												
			}

            if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v1') {

                if ($payment_info['state'] == 'approved') {

                    $this->shop->view['wpsg_mod_paypalapi']['done'] = '1';

                } else {

                    $json_data = $this->api_startPayment($done_view['basket'], $order_id);

                    $approval_url = "";
                    foreach ($json_data['links'] as $k => $l) if ($l['rel'] == 'approval_url') $approval_url = $l['href'];

                    $this->shop->view['wpsg_mod_paypalapi']['approval_url'] = $approval_url;
                    $this->shop->view['wpsg_mod_paypalapi']['mode'] = (($this->isSandbox())?'sandbox':'live');

                }

            } else {

 			    if (isset($payment_info['status']) && $payment_info['status'] == 'COMPLETED') {
			
				$this->shop->view['wpsg_mod_paypalapi']['done'] = '1';				

                } else {

                    $json_data = $this->api_startPayment_v2($done_view['basket'], $order_id);

                    $approval_url = "";
                    foreach ($json_data['links'] as $k => $l) if ($l['rel'] == 'approve' || $l['rel'] == 'approval_url') $approval_url = $l['href'];

                    $this->shop->view['wpsg_mod_paypalapi']['approval_url'] = $approval_url;
                    $this->shop->view['wpsg_mod_paypalapi']['mode'] = (($this->isSandbox())?'sandbox':'live');

                }

            }

			$this->shop->render(WPSG_PATH_VIEW.'/mods/mod_paypalapi/order_done.phtml');
		
		} // public function order_done(&$order_id)

		public function order_view_afterpayment(&$order_id)
		{
			
			$oOrder = wpsg_order::getInstance($order_id);
			
			if ($oOrder->getPaymentID() == $this->id)
			{
				 
				$this->shop->render(WPSG_PATH_VIEW.'/mods/mod_paypalapi/order_view_afterpayment.phtml');
				
			}
			
		}
		
		public function basket_inner_prebutton(&$basket_view) 
		{ 

			// PayPal überhaupt möglich
			if (!array_key_exists($this->id, $this->shop->arPayment)) return;
			 
			// Bestellungen mit 0,- müssen nicht bearbeitet werden
			if ($basket_view['basket']['sum']['preis_gesamt_brutto'] <= 0) return;
									
			// PayPal oder PayPal Express ist deaktiviert
			if (!wpsg_isSizedInt($this->shop->get_option('wpsg_mod_paypalapi_aktiv')) || !wpsg_isSizedInt($this->shop->get_option('wpsg_mod_paypalapi_paypalexpress'))) return false;
			
			if (wpsg_isSizedInt($this->shop->get_option('wpsg_mod_paypalapi_plus_basket'))) {
				
				$this->setViewForPayPalPLus();
				$this->shop->view['wpsg_mod_paypalapi']['disableContinue'] = 'wpsg_basket_submit';
				$this->shop->view['wpsg_mod_paypalapi']['enableContinue'] = 'wpsg_basket_submit';
				
			}
			
			$this->shop->render(WPSG_PATH_VIEW.'/mods/mod_paypalapi/basket_inner_prebutton.phtml');
			
		} // public function basket_inner_prebutton(&$basket_view)
				
		public function basket_save_done_array(&$order_id, &$kunde_id, &$arBasket) 
		{ 
			
			if (wpsg_isSizedString($_SESSION['wpsg']['checkout']['paymentId']))
			{
				
				$oOrder = wpsg_order::getInstance($order_id);
				
				$oOrder->setMeta('wpsg_mod_paypalapi_paymentid', $_SESSION['wpsg']['checkout']['paymentId']);
				$oOrder->setMeta('wpsg_mod_paypalapi_payer_id', $_SESSION['wpsg']['checkout']['payer_id']);

			}
			
		} // public function basket_save_done_array(&$order_id, &$kunde_id, &$arBasket)
		
		public function basket_submitSuccess() 
		{ 
			
			// Zahlungsart auf PayPal API festlegen
			$pre_payment = $_SESSION['wpsg']['checkout']['payment'];
			if ($_SESSION['wpsg']['checkout']['payment'] != $this->id) $_SESSION['wpsg']['checkout']['payment'] = $this->id;
			
			$this->shop->basket->initFromSession();
			
			$basketError = false;
			
			if ($this->shop->callMods('basket_check') == false) {
			
				// Hier stimmt was mit den Daten im Warenkorb nicht -> Zurück zum Warenkorb leiten
				// Zum Beispiel Produktvariablen im Warenkorb
				$this->shop->redirect($this->shop->getUrl(wpsg_ShopController::URL_BASKET));
				$basketError = true;
					
			}	
			
			if (!wpsg_isSizedString($_SESSION['wpsg']['checkout']['shipping'])) {
			
				$this->shop->addFrontendError(__('Bitte eine gültige Versandart auswählen.', 'wpsg'));
				$basketError = true;

			}
			
			if (!wpsg_isSizedInt($_SESSION['wpsg']['checkout']['land'])) {
				
				$this->shop->addFrontendError(__('Bitte ein Land auswählen, damit die Versandkosten bestimmt werden können.', 'wpsg'));
				$basketError = true;
				
			}
			
			if ($this->shop->hasMod('wpsg_mod_willcollect')) {
			
				$arCheckout = array(
					'payment' => $_SESSION['wpsg']['checkout']['payment'],
					'shipping' => $_SESSION['wpsg']['checkout']['shipping']
				);
				
				$error = false; $step = 2;
				$this->shop->callMod('wpsg_mod_willcollect', 'checkCheckout', array(&$step, &$error, &$arCheckout));
				
				if ($error === true) 
				{
					
					$basketError = true;
					$_SESSION['wpsg']['checkout']['payment'] = $pre_payment; 
					
				}
				
			}
			 
			// Bei Fehlern zurück zum Warenkorb leiten
			if ($basketError === true) $this->shop->redirect($this->shop->getUrl(wpsg_ShopController::URL_BASKET));
						
			// Zahlung starten
			$this->shop->redirect($this->shop->getUrl(wpsg_ShopController::URL_BASKET, 'wpsg_mod_paypalapi', 'startPayPalExpress'));
			
		} // public function basket_submitSuccess()
		
		/* Redirects */
		
		/**
		 * Wird bei einem WebHook Event aufgerufen
		 * @throws Exception
		 */
		public function webHookRedirect() {
			
			//  
			$this->shop->addShipPay();
			
			$bodyReceived = file_get_contents('php://input');
						
			// Log BEGIN
			if ($this->shop->get_option('wpsg_mod_paypalapi_webhook_log') == '1') {
			 
				$log_file = $this->shop->get_option('wpsg_mod_paypalapi_webhook_logfile');
				
				$log_JsonData = json_decode($bodyReceived, true);
				
				if ($log_JsonData['event_type'] === 'PAYMENT.CAPTURE.COMPLETED') {
	
					$log_paymentID = $log_JsonData['resource']['id'];
					$log_orderID = intval($this->db->fetchOne("SELECT `target_id` FROM `".WPSG_TBL_META."` WHERE `meta_table` = 'WPSG_TBL_ORDER' AND `meta_key` = 'wpsg_mod_paypalapi_saleid' AND `meta_value` = '".wpsg_q($log_paymentID)."' "));
					
				} else {
	
					$log_paymentID = $log_JsonData['resource']['parent_payment'];
					$log_orderID = $this->db->fetchOne("SELECT `target_id` FROM `".WPSG_TBL_META."` WHERE `meta_table` = 'WPSG_TBL_ORDER' AND `meta_key` = 'wpsg_mod_paypalapi_paymentid' AND `meta_value` = '".wpsg_q($log_paymentID)."' ");
	
				}
				
				$strLog = date('d.m.Y H:i:s', time())."\r\nRAW:".$bodyReceived."\r\nDECODED:".print_r($log_JsonData, 1)."\r\nPAYMENT_ID:".$log_paymentID."\r\nGEFUNDENE BESTELLID:".$log_orderID."\r\n---------------\r\n\r\n";
				file_put_contents($log_file, file_get_contents($log_file)."\r\n".$strLog);
			
			}
			// Log END
			
			$jsonData = json_decode($bodyReceived, true);

			if ($jsonData['event_type'] === 'PAYMENT.CAPTURE.COMPLETED') {

			    $saleID = $jsonData['resource']['id'];
			    $order_id = intval($this->db->fetchOne("SELECT `target_id` FROM `".WPSG_TBL_META."` WHERE `meta_table` = 'WPSG_TBL_ORDER' AND `meta_key` = 'wpsg_mod_paypalapi_saleid' AND `meta_value` = '".wpsg_q($saleID)."' "));
			    $amount = $jsonData['resource']['amount']['value'];

            } else {

                $payment_id = $jsonData['resource']['parent_payment'];
                $amount = $jsonData['resource']['amount']['total'];
                $order_id = $this->db->fetchOne("SELECT `target_id` FROM `".WPSG_TBL_META."` WHERE `meta_table` = 'WPSG_TBL_ORDER' AND `meta_key` = 'wpsg_mod_paypalapi_paymentid' AND `meta_value` = '".wpsg_q($payment_id)."' ");

            }

            $oOrder = wpsg_order::getInstance($order_id);
            if ($oOrder === false) die(__('Keine Bestellung zu PaymentID gefunden.', 'wpsg'));

            // Ich ignoriere Zahlungen die bereits extern gezahlt gemeldet wurden
            if ($oOrder->__get('ext_payed')  === '1') exit;

            if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v1') {

                list($payment_state, $sale_state) = $this->getState($order_id);
                $orderPaymentID = $oOrder->getMeta('wpsg_mod_paypalapi_paymentid');

                if ($this->shop->setPayMent($oOrder->id, $amount) && $payment_state === 'approved' && $sale_state === 'completed') {

                    $oOrder->log(__('PayPalAPI V1 WebHook Erfolg', 'wpsg'), print_r($jsonData, 1));

//					try {
                    
						$this->shop->setOrderStatus($oOrder->id, 100, true);
/*
					} catch (\Exception $e) {
						
						mail("daniel@maennchen1.de", "TIERPARK Fehler:".$e->getMessage(), print_r(debug_backtrace(), true));
						
					}*/
					
                    $this->db->UpdateQuery(WPSG_TBL_ORDER, ['ext_payed' => '1'], " `id` = '".wpsg_q($oOrder->getId())."' ");

                } else {

                    $oOrder->log(__('PayPalAPI Abgelehnt', 'wpsg'), print_r($jsonData, 1)."\r\nOrder PaymentID:".$orderPaymentID."\r\nRequest PaymentID:".$payment_id);

                }

            } else if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v2') {

                list($payment_state, $sale_state, $result) = $this->getState($order_id);
                $orderPaymentID = $oOrder->getMeta('wpsg_mod_paypalapi_paymentid');

                if ($this->shop->setPayMent($oOrder->id, $amount) && $payment_state === 'COMPLETED' && $sale_state === 'COMPLETED') {

                    $oOrder->log(__('PayPal Status', 'wpsg'), print_r($result, true));
                    $oOrder->log(__('PayPalAPI V2 WebHook Erfolg', 'wpsg'), print_r($jsonData, 1));

                    $this->shop->setOrderStatus($oOrder->id, 100, true);
                    $this->db->UpdateQuery(WPSG_TBL_ORDER, ['ext_payed' => '1'], " `id` = '".wpsg_q($oOrder->getId())."' ");

                } else {

                    $oOrder->log(__('PayPalAPI Abgelehnt', 'wpsg'), print_r($jsonData, 1)."\r\nOrder PaymentID:".$orderPaymentID."\r\nRequest PaymentID:".$payment_id);

                }

            }

			header('HTTP/1.1 200 OK'); exit(1);
						
		} // public function webHookRedirect()
				
		/**
		 * Wird aus dem Warenkorb bei PayPal Express aufgerufen
		 */
		public function startPayPalExpressRedirect() {#
		 
			//Bestellung wird im BasketController schon angelegt
			//$order_id = $this->shop->basket->save(false);
			$order_id = $_SESSION['wpsg']['order_id'];
						
			$json_data = $this->api_startPayment_v2($this->shop->basket->toArray(), $order_id);

			$approval_url = false;
			foreach ($json_data['links'] as $l) { if ($l['rel'] == 'approve' || $l['rel'] == 'approval_url') $approval_url = $l['href']; }

			$this->shop->redirect($approval_url);
			
		} // public function startPayPalExpressRedirect()
		 
		/**
		 * Wird aufgerufen wenn der Kunde von PayPal zurück zum Shop kommt
		 */
		public function returnPayPalExpressRedirect() {

            if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v1') {

                $json_data = $this->api_getPaymentInfo($_REQUEST['paymentId']);

            } else {

                $_REQUEST['paymentId'] = sanitize_text_field($_REQUEST['token']);

			    $json_data = $this->api_getPaymentInfo_v2($_REQUEST['paymentId']);

            }

			$order_id = $this->db->fetchOne("SELECT `target_id` FROM `".WPSG_TBL_META."` WHERE `meta_table` = 'WPSG_TBL_ORDER' AND `meta_key` = 'wpsg_mod_paypalapi_paymentid' AND `meta_value` = '".wpsg_q($_REQUEST['paymentId'])."' ");

			if (wpsg_isSizedInt($order_id)) {
				
				$oOrder = wpsg_order::getInstance($order_id);
				
				if ($oOrder !== false && $oOrder->status != wpsg_ShopController::STATUS_UNVOLLSTAENDIG) {
						
					$json_data = $this->api_executePayment_v2($_REQUEST['paymentId'], $_REQUEST['PayerID']);
					
					$oOrder->setMeta('wpsg_mod_paypalapi_saleid', $json_data['purchase_units'][0]['payments']['captures'][0]['id']);
					
					$this->getState($oOrder->id);
					
					if (wpsg_isSizedString($json_data['transactions'][0]['related_resources'][0]['sale']['id'])) $oOrder->setMeta('wpsg_mod_paypalapi_saleid', $json_data['transactions'][0]['related_resources'][0]['sale']['id']);
					
					$this->shop->redirect($this->shop->getDoneURL($oOrder->id));
						
				}

			}

			$_SESSION['wpsg']['checkout']['payment'] = $this->id;

            // billing_address scheint es nicht mehr zu geben (Aktuell steht es aber noch in der Doku)
            // In der Doku steht aber auch, dass shipping_address DEPRECATED ist :?
            // Wir haben hier billing_address durch shipping_address ersetzt
            // https://developer.paypal.com/docs/api/payments/#payment_get
            // https://stackoverflow.com/questions/44968561/how-to-get-billing-address-in-getexpresscheckoutdetails-api-nvp

			if (!wpsg_isSizedString($_SESSION['wpsg']['checkout']['vname']) && isset($json_data['payer']['name']['given_name'])) $_SESSION['wpsg']['checkout']['vname'] = $json_data['payer']['name']['given_name'];
			if (!wpsg_isSizedString($_SESSION['wpsg']['checkout']['name']) && isset($json_data['payer']['name']['surname'])) $_SESSION['wpsg']['checkout']['name'] = $json_data['payer']['name']['surname'];
			if (!wpsg_isSizedString($_SESSION['wpsg']['checkout']['email']) && isset($json_data['payer']['email_address'])) {
				$_SESSION['wpsg']['checkout']['email'] = $json_data['payer']['email_address'];
				$_SESSION['wpsg']['checkout']['email2'] = $json_data['payer']['email_address'];
			}				
			
			if (!wpsg_isSizedString($_SESSION['wpsg']['checkout']['tel']) && isset($json_data['payer']['phone'])) {
				
				$request_tel = null;
				
				if (is_array($json_data['payer']['phone']) && isset( $json_data['payer']['phone']['phone_number']['national_number'])) {
			
					$request_tel =  $json_data['payer']['phone']['phone_number']['national_number'];
					
				} else if (is_string($json_data['payer']['phone'])) {
				
					$request_tel = $json_data['payer']['phone'];
					
				}
				
				if ($request_tel !== null) $_SESSION['wpsg']['checkout']['tel'] = $json_data['payer']['phone'];
				
			}
			
			if (!wpsg_isSizedString($_SESSION['wpsg']['checkout']['strasse']) && isset($json_data['payer']['address']['address_line_1'])) $_SESSION['wpsg']['checkout']['strasse'] =  $json_data['payer']['address']['address_line_1'];
			if (!wpsg_isSizedString($_SESSION['wpsg']['checkout']['ort']) && isset($json_data['payer']['address']['admin_area_2'])) $_SESSION['wpsg']['checkout']['ort'] = $json_data['payer']['address']['admin_area_2'];
			if (!wpsg_isSizedString($_SESSION['wpsg']['checkout']['plz']) && isset($json_data['payer']['address']['postal_code'])) $_SESSION['wpsg']['checkout']['plz'] = $json_data['payer']['address']['postal_code'];
			if (!wpsg_isSizedString($_SESSION['wpsg']['checkout']['land']) && isset($json_data['payer']['address']['country_code'])) $_SESSION['wpsg']['checkout']['land'] = $this->db->fetchOne("SELECT `id` FROM `".WPSG_TBL_LAND."` WHERE UPPER(`kuerzel`) = '".wpsg_q(strtoupper($json_data['payer']['address']['country_code']))."'");

			if ($this->shop->hasMod('wpsg_mod_shippingadress') && !isset($_SESSION['wpsg']['checkout']['diff_shippingadress'])) {

                $shipping = wpsg_getArray($json_data['purchase_units'][0]['shipping']);

			    if (wpsg_isSizedArray($shipping)) {

                    $country_id = intval($this->db->fetchOne("SELECT `id` FROM `".WPSG_TBL_LAND."` WHERE UPPER(`kuerzel`) = '".wpsg_q(strtoupper(wpsg_getStr($shipping['address']['country_code'])))."'"));

					if ($country_id > 0) {
						
						$_SESSION['wpsg']['checkout']['diff_shippingadress'] = '1';
						
						$arName = wpsg_explodeName($shipping['name']['full_name']);
						
						$_SESSION['wpsg']['checkout']['shipping_vname'] = wpsg_getStr($arName[0]);
						$_SESSION['wpsg']['checkout']['shipping_name'] = wpsg_getStr($arName[1]);
						
						$_SESSION['wpsg']['checkout']['shipping_strasse'] = wpsg_getStr($shipping['address']['address_line_1']);
						$_SESSION['wpsg']['checkout']['shipping_plz'] = wpsg_getStr($shipping['address']['postal_code']);
						$_SESSION['wpsg']['checkout']['shipping_ort'] = wpsg_getStr($shipping['address']['admin_area_2']);
						
						$_SESSION['wpsg']['checkout']['shipping_land'] = $country_id;
						
					}

                }

            }

			$_SESSION['wpsg']['checkout']['payment_amount'] = $json_data['purchase_units'][0]['amount']['value'];
			$_SESSION['wpsg']['checkout']['paymentId'] = $json_data['id'];

            if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v1') {

                $_SESSION['wpsg']['checkout']['payer_id'] = $json_data['payer']['payer_info']['payer_id'];

            } else {

			    $_SESSION['wpsg']['checkout']['payer_id'] = $json_data['payer']['payer_id'];

            }
			
			$this->shop->basket->initFromSession(true);
			
			// Hier ist das Speichern ok. Dadurch werden die Neuen Kundendaten von PayPal in der Unvollständigen Bestellung aktualisiert
			$this->shop->basket->save(false);
									
			$check1 = $checkCheckout = $this->shop->basket->checkCheckout(1);
			$check2 = $checkCheckout = $this->shop->basket->checkCheckout(2);
			
			if ($check1 === false) {
					
				$this->shop->redirect($this->shop->getUrl(wpsg_ShopController::URL_CHECKOUT));
					
			} else if ($check2 === false) {
					
				$this->shop->redirect($this->shop->getUrl(wpsg_ShopController::URL_CHECKOUT2));
					
			} else {
			
				$this->shop->addFrontendMessage(__('Ihre Daten wurden von PayPal übernommen und Sie können die Bestellung direkt abschließen, nachdem Sie alles überprüft haben.', 'wpsg'));
					
				$this->shop->redirect($this->shop->getUrl(wpsg_ShopController::URL_OVERVIEW));
					
			}
			
		}
		 
		/**
		 * Wird für den AutoRefresh der Versandarten bei PayPal PLUS verwendet 
		 */
		public function ShippingSelectRedirect()
		{
			
			$_SESSION['wpsg']['checkout']['shipping'] = $_REQUEST['wpsg']['checkout']['shipping'];
				
			$this->shop->redirect($this->shop->getUrl(wpsg_ShopController::URL_CHECKOUT2));
			
		} // public function ShippingSelectRedirect()
		
		/**
		 * Wird bei PayPal Plus aufgerufen wenn eine Zahlungsmethode aus dem Shop aufgerufen wurde
		 */
		public function PaymentSelectRedirect()
		{
			
			$_SESSION['wpsg']['checkout']['payment'] = wpsg_sinput("key", $_REQUEST['payment_id']);
			
			$this->shop->redirect($this->shop->getUrl(wpsg_ShopController::URL_OVERVIEW));
						
		} // public function PaymentSelectRedirect()
		
		/* Aktionen */
		
		/* Modulfunktionen */
		
		/**
		 * Definiert Variablen für das View welches den IFrame einbindet
		 */
		private function setViewForPayPalPLus() {
			
			$json_data = $this->api_startPayment_v2($this->shop->view['basket'], $_SESSION['wpsg']['order_id'], false);

			//wpsg_debug($json_data);

			$approval_url = false;
			foreach ($json_data['links'] as $l) { if ($l['rel'] == 'approve' || $l['rel'] == 'approval_url') $approval_url = $l['href']; }
			
			// Das war mal drin mir ist nicht klar warum:
			//$_SESSION['wpsg']['checkout']['payment'] = false;
			
			$arThirdPartyPayment = array();
			
			foreach ($this->shop->arPayment as $payment_id => $payment)
			{
				
				if (strval($payment_id) != strval($this->id))
				{
					 
					$strPricePrefix = '';
					
					if ($payment['price'] > 0) 
					{
						
						$strPricePrefix = wpsg_translate(__('Bei Auswahl dieser Zahlungsart entstehen Kosten von #1#.', 'wpsg'), wpsg_ff($payment['price'], $this->shop->get_option('wpsg_currency'))).' ';
						
					}
					
					$tpp = array(
						'redirectUrl' => $this->shop->getUrl(wpsg_ShopController::URL_BASKET, 'wpsg_mod_paypalapi', 'PaymentSelect', array('payment_id' => $payment_id)),
						'methodName' => $payment['name'],
						'description' => $strPricePrefix.$payment['hint']
					);
					
					if (wpsg_isSizedString($payment['logo'])) $tpp['imageUrl'] = $payment['logo'];
					 
					$arThirdPartyPayment[] = $tpp;
					
				}
				
			}
			 
			$this->shop->view['wpsg_mod_paypalapi']['ThirdPartyPayment'] = $arThirdPartyPayment;
			$this->shop->view['wpsg_mod_paypalapi']['mode'] = (($this->isSandbox())?'sandbox':'live');
			$this->shop->view['wpsg_mod_paypalapi']['approval_url'] = $approval_url;
			
			$this->shop->view['wpsg_mod_paypalapi']['disableContinue'] = 'wpsg_checkout2button';
			$this->shop->view['wpsg_mod_paypalapi']['enableContinue'] = 'wpsg_checkout2button';			
			
		}
		
		/**
		 * Wird optional beim speichern ausgelöst, versucht den WebHook anzulegen und setzt die BackendMeldungen
		 */
		private function createWebHook() {

			$create_webhook = false; $nExists = false;
			$wh_url = $this->shop->getUrl(wpsg_ShopController::URL_BASKET, 'wpsg_mod_paypalapi', 'webHook', array(), true);

			$arWH = $this->api_getWebHooks();

			if (wpsg_isSizedArray($arWH)) {

				foreach ($arWH as $wh) {
				
					if ($wh['url'] === $wh_url) {

						$nExists = true;
						break;
				
					}
				
				}
				
				if ($nExists === false) $create_webhook = true;
				else {
						
					$this->shop->addBackendError(__('WebHook ist bereits registriert.', 'wpsg'));
						
				}
				
			} else $create_webhook = true;

			if ($create_webhook === true) {

				$bOK = $this->api_createWebHook($wh_url);

				if ($bOK === true) $this->shop->addBackendMessage(__('WebHook erfolgreich angelegt.', 'wpsg'));
				else $this->shop->addBackendError(__('WebHook konnte nicht angelegt werden.', 'wpsg'));
				
			}

		} // private function createWebHook();
		
		/**
		 * Gibt den PaymentState und SaleState zurück
		 */
		public function getState($order_id) {
												
			$oOrder = wpsg_order::getInstance($order_id);

			$paymentID = $oOrder->getMeta('wpsg_mod_paypalapi_paymentid');

			if (!wpsg_isSizedString($paymentID)) return false;
			else {

                if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v1') {

                    $payment_info = $this->api_getPaymentInfo($oOrder->getMeta('wpsg_mod_paypalapi_paymentid'));

                    $payment_state = $payment_info['state'];
                    $sale_state = $payment_info['transactions'][0]['related_resources'][0]['sale']['state'];

                    if (!wpsg_isSizedString($payment_state)) return false;

                    if (wpsg_isSizedString($payment_state)) $oOrder->setMeta('wpsg_mod_paypalapi_paymentstate', $payment_state);
                    if (wpsg_isSizedString($sale_state)) $oOrder->setMeta('wpsg_mod_paypalapi_salestate', $sale_state);

                    return array($payment_state, $sale_state, $payment_info);

                } else {

                    $payment_info = $this->api_getPaymentInfo_v2($paymentID);

                    $payment_state = wpsg_getStr($payment_info['status']);
                    $sale_state = wpsg_getStr($payment_info['purchase_units'][0]['payments']['captures'][0]['status'], __('Unbekannt', 'wpsg'));

                    if (!wpsg_isSizedString($payment_state)) return false;

                    if (wpsg_isSizedString($payment_state)) $oOrder->setMeta('wpsg_mod_paypalapi_paymentstate', $payment_state);
                    if (wpsg_isSizedString($sale_state)) $oOrder->setMeta('wpsg_mod_paypalapi_salestate', $sale_state);

                    return array($payment_state, $sale_state, $payment_info);

                }

            }
			
		} // public function getState($order_id)
		
		public function stornoOrder($order_id) {
			
			$oOrder = wpsg_order::getInstance($order_id);
			if ($oOrder === false) throw new \wpsg\Exception(__('Bei einer Stornierung konnte Bestellung nicht geladen werden', 'wpsg'));

			$payment_id = $oOrder->getMeta('wpsg_mod_paypalapi_paymentid');
			$sale_id = $oOrder->getMeta('wpsg_mod_paypalapi_saleid');

            $payment_info = $this->api_getPaymentInfo_v2($payment_id);

			$json_data = $this->api_refundSale_v2($sale_id, $oOrder->getAmount());
			
			$this->getState($oOrder->id);			
			
			return true;
			
		} // public function stornoOrder($order_id)

		/**
		 * Gibt true/false zurück ob die Api auf die PayPal Sandbox zurückgreift
		 * @return boolean
		 */
		public function isSandbox()
		{
			
			if (wpsg_isSizedInt($this->shop->get_option('wpsg_mod_paypalapi_sandbox')))
			{
				
				return true;
				
			}
			else
			{
				
				return false;
				
			}
			
		} // public function isSandbox()
		
		/* Neu ohne PHPApi */

        /**
         * DEPRECATED
         * @param $sale_id
         * @param $amount
         * @return array|mixed|void
         */
		private function api_refundSale($sale_id, $amount) {

			$post_data = array(
				'amount' => array(
					'total' => number_format($amount, 2, '.', ''),
					'currency' => $this->currency
				)		
			);
			
			$json_data = $this->api_call('v1/payments/sale/'.$sale_id.'/refund', $post_data, $this->api_getAccessToken());
			
			return $json_data;
			
		} // private function api_refundSale($sale_id)

        private function api_refundSale_v2($sale_id, $amount) {

            if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v1') {

                return $this->api_refundSale($sale_id, $amount);

            }

            $client = new \PayPalCheckoutSdk\Core\PayPalHttpClient($this->api_getEnvironment());

            $request = new \PayPalCheckoutSdk\Payments\CapturesRefundRequest($sale_id);
            $request->body = [
                'amount' => [
                    'value' => number_format($amount, 2, '.', ''),
                    'currency_code' => $this->currency
                ]
            ];

            $json_data = $client->execute($request);
            $json_data = json_decode(json_encode($json_data), true)['result'];

            return $json_data;

        }

		/**
		 * Erstellt einen neuen WebHook für Zahlungseingänge auf der übergebenene URL
		 */
		private function api_createWebHook($url) {
			
			$post_data = array(
				'url' => $url,
				'event_types' => array(
					[
						'name' => 'PAYMENT.SALE.COMPLETED'
					], [
					    'name' => 'PAYMENT.CAPTURE.COMPLETED'
                    ]
				)
			);
			
			$json_data = $this->api_call('v1/notifications/webhooks', $post_data, $this->api_getAccessToken());

			if (wpsg_isSizedString($json_data['id'])) return true;
			else return false;
			
		} // private function api_createWebHook($url)
		
		/**
		 * Gibt alle registrierten WebHooks zurück
		 */
		private function api_getWebHooks()
		{
			
			$json_data = $this->api_call('v1/notifications/webhooks', false, $this->api_getAccessToken());
			
			return $json_data['webhooks'];
			
		} // private function api_getWebHooks()
		
		/**
         * DEPRECATED
		 * Führt eine Zahlung aus
		 */
		private function api_executePayment($payment_id, $payer_id) {

			$post_data = array(
				'payer_id' => $payer_id
			);
			
			$json_data = $this->api_call('v1/payments/payment/'.$payment_id.'/execute', $post_data, $this->api_getAccessToken());

			return $json_data;
			
		} // private function api_executePayment($payment_id, $payer_id, $order_id)

        private function api_executePayment_v2($payment_id, $payer_id) {

            if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v1') {

                return $this->api_executePayment($payment_id, $payer_id);

            }

            $client = new \PayPalCheckoutSdk\Core\PayPalHttpClient($this->api_getEnvironment());
            $json_data = $client->execute(new \PayPalCheckoutSdk\Orders\OrdersCaptureRequest($payment_id));
            $json_data = json_decode(json_encode($json_data), true)['result'];

            return $json_data;

        }

		/**
		 * Aktualisiert eine Zahlungserfahrung
		 */
		private function api_updateWebExperience($paymentExperienceID)
		{
			
			$brand_name = $this->shop->replaceUniversalPlatzhalter(__($this->shop->get_option('wpsg_mod_paypalapi_experience_label'), 'wpsg'));
			$logo_image = $this->shop->get_option('wpsg_mod_paypalapi_experience_logo');
			
			$post_data = array(
				'name' => 'wpShopGermany',
				'presentation' => array(					
					'locale_code' => $this->shop->get_option('wpsg_mod_paypalapi_experience_locale_code')
				)
			);
			
			if (wpsg_isSizedString($brand_name)) $post_data['presentation']['brand_name'] = $brand_name;
			if (wpsg_isSizedString($logo_image)) $post_data['presentation']['logo_image'] = $logo_image;			
			
			$json_data = $this->api_call('v1/payment-experience/web-profiles/'.$paymentExperienceID, $post_data, $this->api_getAccessToken(), false, true, true);
				
			return $json_data;
			
		} // private function api_updateWebExperience($paymentExperienceID)
		
		/**
		 * Erstellt eine Zahlungserfahrung
		 */
		private function api_createWebExperience()
		{
			 
			$brand_name = $this->shop->replaceUniversalPlatzhalter(__($this->shop->get_option('wpsg_mod_paypalapi_experience_label'), 'wpsg'));
			$logo_image = $this->shop->get_option('wpsg_mod_paypalapi_experience_logo');
			
			$post_data = array(
				'name' => 'wpShopGermany',
				'presentation' => array(
					'locale_code' => $this->shop->get_option('wpsg_mod_paypalapi_experience_locale_code')
				)
			);
			
			if (wpsg_isSizedString($brand_name)) $post_data['presentation']['brand_name'] = $brand_name;
			if (wpsg_isSizedString($logo_image)) $post_data['presentation']['logo_image'] = $logo_image;
			
			$json_data = $this->api_call('v1/payment-experience/web-profiles', $post_data, $this->api_getAccessToken());
				
			return $json_data;
			
		} // private function api_createWebExperience()
		
		/**
		 * Gibt Informationen über angelegte Zahlungserfahrungen zurück
		 */
		private function api_getWebExperience()
		{
			
			$access_token = $this->api_getAccessToken();
			
			$json_data = $this->api_call('v1/payment-experience/web-profiles', false, $access_token);
			
			return $json_data;
			
		} // private function api_getWebExperience()

        private function api_getEnvironment() {

            if ($this->isSandbox()) {

                $clientId = $this->shop->get_option('wpsg_mod_paypalapi_sandbox_clientid');
                $clientSecret = $this->shop->get_option('wpsg_mod_paypalapi_sandbox_secret');

                return new \PayPalCheckoutSdk\Core\SandboxEnvironment($clientId, $clientSecret);

            } else {

                $clientId = $this->shop->get_option('wpsg_mod_paypalapi_clientid');
                $clientSecret = $this->shop->get_option('wpsg_mod_paypalapi_secret');

                return new \PayPalCheckoutSdk\Core\ProductionEnvironment($clientId, $clientSecret);

            }

        }

        /**
         * Authorisiert eine Zahlung
         * https://developer.paypal.com/docs/api/payments/v1/#definition-transaction
         * https://developer.paypal.com/docs/api/payments/v1/#definition-payer
         *
         */
        private function api_startPayment($arBasket, $order_id = false, $express = true) {

            // Dadurch wird die Bestellung in der DB fixiert und Platzhalter kännen generiert werden
            if (!wpsg_isSizedInt($order_id)) $order_id = $this->shop->basket->save(false, false, true);

            $oOrder = wpsg_order::getInstance($order_id);
            $access_token = $this->api_getAccessToken();

            $item_list = array(
                'items' => array()
            );

            $post_data = array();

            $post_data['transactions'] = array(
                array(
                    'amount' => array(
                        'total' => number_format($arBasket['arCalculation']['sum']['topay_brutto'], 2, '.', ''),
                        'currency' => $this->currency
                    ),
                    'description' => $this->shop->replaceUniversalPlatzhalter(__($this->shop->get_option('wpsg_mod_paypalapi_paypalexpress_subject'), 'wpsg'), $order_id),
                )
            );

            $subTotal = 0;

            // Produkte
            foreach ($arBasket['produkte'] as $k => $p)
            {

                $product_price = ((wpsg_isSizedString($arBasket['noMwSt'], '1'))?$p['preis_netto']:$p['preis_brutto']);

                $sku = $this->shop->getProductAnr($this->shop->getProduktID($p['id']));
                if (!wpsg_isSizedString($sku)) $sku = __('n.A.', 'wpsg');

                $item_list['items'][] = array(
                    'quantity' => $p['menge'],
                    'name' => $this->shop->getProductName($this->shop->getProduktID($p['id']), true),
                    'description' => '',
                    'sku' => $sku,
                    'url' => $this->shop->getProduktLink($this->shop->getProduktID($p['id'])),
                    'price' => number_format($product_price, 2, '.', ''),
                    'currency' => $this->currency
                );

                $subTotal += $product_price * $p['menge'];

            }

			// Wertgutschein - funktioniert von PayPal Seite her aus nicht, siehe Kommentar weiter unten
			/*
				if (isset($arBasket['arCalculation']['coupon'])) {

						$item_list['items'][] = array(
							'quantity' => '1',
							'name' => __('Wertgutschein', 'wpsg'),
							'description' => wpsg_translate(__('Eingelöster Gutschein Code:#1#', 'wpsg'), $arBasket[arCalculation]['coupon']['code']),
							'sku' => $sku,
							//'price' => -number_format($arBasket['arCalculation']['coupon']['netto'], 2, '.', ''),
							'currency' => $this->currency,
						);

						$subTotal += -$arBasket['arCalculation']['coupon']['brutto'];

					} 
			*/

            // Gutschein
            if (isset($arBasket['arCalculation']['voucher']['brutto']))
            {

                $item_list['items'][] = array(
                    'quantity' => '1',
                    'name' => __('Gutschein', 'wpsg'),
                    'description' => wpsg_translate(__('Eingelöster Gutschein Code:#1#', 'wpsg'), $arBasket['arCalculation']['voucher']['code']),
                    //'price' => number_format(-1 * $arBasket['sum']['gs_brutto'], 2, '.', ''),
                    'discount' => $arBasket['arCalculation']['voucher']['brutto'],
					'currency' => $this->currency
                );

                $subTotal += -1 * $arBasket['arCalculation']['voucher']['brutto'];

            }
			

            // Zahlungskosten/Vergünstigungen
            if (isset($arBasket['arCalculation']['sum']['payment_brutto']) && $arBasket['arCalculation']['sum']['payment_brutto'] != 0)
            {

                $item_list['items'][] = array(
                    'quantity' => '1',
                    'name' => (($arBasket['arCalculation']['sum']['payment_brutto'] > 0)?__('Gebühr für Zahlungsart', 'wpsg'):__('Rabatt durch Zahlungsart', 'wpsg')),
                    'description' => __('Rabatt, der Aufgrund der gewählten Zahlungsart entstanden ist.', 'wpsg'),
                    'price' => number_format($arBasket['arCalculation']['sum']['payment_brutto'], 2, '.', ''),
                    'currency' => $this->currency
                );

                $subTotal += $arBasket['arCalculation']['sum']['payment_brutto'];

            }

            // Rabatte auf Warenkorb
            if (isset($arBasket['sum']['preis_rabatt_brutto']) && $arBasket['sum']['preis_rabatt_brutto'] > 0)
            {

                $item_list['items'][] = array(
                    'quantity' => '1',
                    'name' => __('Rabatt', 'wpsg'),
                    'description' => __('Rabatt, der sich aus dem Warenkorbwert ergibt.', 'wpsg'),
                    'price' => number_format(-1 * $arBasket['sum']['preis_rabatt_brutto'], 2, '.', ''),
                    'currency' => $this->currency
                );

                $subTotal += -1 * $arBasket['sum']['preis_rabatt_brutto'];

            }

            /** @var wpsg_country $oInvoiceCountry */
            $oInvoiceCountry = null; if (wpsg_isSizedInt($arBasket['checkout']['land'])) $oInvoiceCountry = wpsg_country::getInstance($arBasket['checkout']['land']);

            /** @var wpsg_country $oShippingCountry */
            $oShippingCountry = null;
            if ($oInvoiceCountry !== null) $oShippingCountry = $oInvoiceCountry;
            if (wpsg_isSizedInt($arBasket['checkout']['shipping_land'])) $oShippingCountry = wpsg_country::getInstance($arBasket['checkout']['shipping_land']);

            $post_data['intent'] = 'sale';
            $post_data['payer'] = array(
                'payment_method' => 'paypal'
            );

            if ($oInvoiceCountry !== null && wpsg_isSizedString($arBasket['checkout']['plz'])) {

                $post_data['payer']['payer_info'] = [
                    'email' => $arBasket['checkout']['email'],
                    'first_name' => $arBasket['checkout']['vname'],
                    'last_name' => $arBasket['checkout']['name'],
                    'billing_address' => [
                        'line1' => trim($arBasket['checkout']['strasse']),
                        'line2' => '',
                        'city' => trim($arBasket['checkout']['ort']),
                        'country_code' => $oInvoiceCountry->getShorttext(),
                        'postal_code' => trim($arBasket['checkout']['plz']),
                    ]
                ];

                if (strtotime($arBasket['checkout']['geb']) > 0) {

                    $post_data['payer']['payer_info']['birth_date'] = date('Y-m-d', strtotime($arBasket['checkout']['geb']));

                }

                if (preg_match('/^(?:\((\+?\d+)?\)|\+?\d+) ?\d*(-?\d{2,3} ?){0,4}$/', $arBasket['checkout']['tel'])) {

                    $post_data['payer']['payer_info']['billing_address']['phone'] = preg_replace('/ /', '', $arBasket['checkout']['tel']);

                }

            }

            if ($oOrder->getId() > 0 && $oOrder->status == wpsg_ShopController::STATUS_UNVOLLSTAENDIG)
            {

                $post_data['redirect_urls'] = array(
                    'return_url' => $this->shop->getUrl(wpsg_ShopController::URL_BASKET, 'wpsg_mod_paypalapi', 'returnPayPalExpress'),
                    'cancel_url' => $this->shop->getUrl(wpsg_ShopController::URL_BASKET)
                );

            }
            else
            {

                $post_data['redirect_urls'] = array(
                    'return_url' => $this->shop->getUrl(wpsg_ShopController::URL_BASKET, 'wpsg_mod_paypalapi', 'returnPayPalExpress'),
                    'cancel_url' => $this->shop->getDoneURL($oOrder->id)
                );

            }

            // Versandkosten/Vergünstigungen
            if (isset($arBasket['arCalculation']['shipping']['brutto']) && $arBasket['arCalculation']['shipping']['brutto'] > 0)
            {

                $details['shipping'] = number_format($arBasket['arCalculation']['shipping']['brutto'], 2, '.', '');

            }
            else if (isset($arBasket['arCalculation']['shipping']['brutto']) && $arBasket['arCalculation']['shipping']['brutto'] < 0)
            {

                $item_list['items'][] = array(
                    'quantity' => '1',
                    'name' => __('Rabatt aufgrund Versandart', 'wpsg'),
                    'price' => number_format($arBasket['arCalculation']['shipping']['brutto'], 2, '.', ''),
                    'currency' => $this->currency
                );

                $subTotal += $arBasket['arCalculation']['shipping']['brutto'];

            }

            //wpsg_debug($item_list);
            //wpsg_debug($subTotal);

            /*
             * Daschmi@14.12.2016
             * Ich vergleiche hier die Summe der Positionen mit dem übergebenen Brutto
             * Bei Nettorechnungen kann es hier zu Problemen führen Siehe Track Ticket 505
             */

            $shipping_address = [];

            // Lieferadresse
            if ($oShippingCountry !== null) {

                if (wpsg_getStr($arBasket['checkout']['diff_shippingadress']) == '1') {

                    $shipping_address = [
                        'recipient_name' => trim($arBasket['checkout']['shipping_firma'].' '.$arBasket['checkout']['shipping_vname'].' '.$arBasket['checkout']['shipping_name']),
                        'line1' => trim($arBasket['checkout']['shipping_strasse']),
                        'line2' => '',
                        'city' => trim($arBasket['checkout']['shipping_ort']),
                        'country_code' => $oShippingCountry->getShorttext(),
                        'postal_code' => trim($arBasket['checkout']['shipping_plz'])
                    ];

                } else {

                    $shipping_address = [
                        'recipient_name' => trim($arBasket['checkout']['firma'].' '.$arBasket['checkout']['vname'].' '.$arBasket['checkout']['name']),
                        'line1' => trim($arBasket['checkout']['strasse']),
                        'line2' => '',
                        'city' => trim($arBasket['checkout']['ort']),
                        'country_code' => $oShippingCountry->getShorttext(),
                        'postal_code' => trim($arBasket['checkout']['plz'])
                    ];

                }

            }

            if (wpsg_isSizedArray($shipping_address) && isset($post_data['transactions'][0]['item_list'])) {

                $post_data['transactions'][0]['item_list']['shipping_address'] = $shipping_address;

            }

            if (wpsg_isSizedInt($this->shop->get_option('wpsg_mod_paypalapi_paypalexpress_details')) && floatval($subTotal) === floatval($arBasket['arCalculation']['sum']['topay_brutto']))
            {

                if (wpsg_isSizedArray($details))
                {

                    $post_data['transactions'][0]['amount']['details'] = $details;
                    $post_data['transactions'][0]['amount']['details']['subtotal'] = number_format($subTotal, 2, '.', '');

                }

                $post_data['transactions'][0]['item_list'] = $item_list;
            }
            else
            {

                $post_data['transactions'][0]['item_list'] = array(
                    'items' => array(
                        array(
                            'quantity' => '1',
                            'name' => $this->shop->replaceUniversalPlatzhalter(__($this->shop->get_option('wpsg_mod_paypalapi_paypalexpress_subject'), 'wpsg'), $order_id),
                            'price' => number_format($arBasket['arCalculation']['sum']['topay_brutto'], 2, '.', ''),
                            'currency' => $this->currency
                        )
                    )
                );

            }

            if (wpsg_isSizedString($this->shop->get_option('wpsg_mod_paypalapi_paymentExperience')))
            {

                $post_data['experience_profile_id'] = $this->shop->get_option('wpsg_mod_paypalapi_paymentExperience');

            }

            //die('<pre>'.print_r($post_data, true).'</pre>');

            $json_data = $this->api_call('v1/payments/payment', $post_data, $access_token);

            if (wpsg_isSizedString($json_data['id'])) {

                $oOrder->setMeta('wpsg_mod_paypalapi_paymentid', $json_data['id']);

            }

            return $json_data;

        } // private function api_startPayment()

        /**
         * @param $arBasket
         * @param bool $order_id
         * @param bool $express
         * @throws \wpsg\Exception
         */
        private function api_startPayment_v2($arBasket, $order_id = false, $express = true) {

            if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v1') {

                return $this->api_startPayment($arBasket, $order_id, $express);

            }

            if (!wpsg_isSizedInt($order_id)) $order_id = $this->shop->basket->save(false, false, true);

            $oOrder = wpsg_order::getInstance($order_id);

            $client = new \PayPalCheckoutSdk\Core\PayPalHttpClient($this->api_getEnvironment());

            if ($oOrder->getId() > 0 && intval($oOrder->status) !== wpsg_ShopController::STATUS_UNVOLLSTAENDIG) {

                $application_context = [
                    'return_url' => $this->shop->getUrl(wpsg_ShopController::URL_BASKET, 'wpsg_mod_paypalapi', 'returnPayPalExpress'),
                    'cancel_url' => $this->shop->getDoneURL($oOrder->id)
                ];

            } else {

                $application_context = [
                    'return_url' => $this->shop->getUrl(wpsg_ShopController::URL_BASKET, 'wpsg_mod_paypalapi', 'returnPayPalExpress'),
                    'cancel_url' => $this->shop->getUrl(wpsg_ShopController::URL_BASKET)
                ];

            }

            $request = new \PayPalCheckoutSdk\Orders\OrdersCreateRequest();
            $request->prefer('return=representation');
            $request->body = [
                "intent" => "CAPTURE",
                'purchase_units' => [[
                    'reference_id' => "order_".$order_id,
                    'description' => $this->shop->replaceUniversalPlatzhalter(__($this->shop->get_option('wpsg_mod_paypalapi_paypalexpress_subject'), 'wpsg'), $order_id),
                    'amount' => [
                        'currency_code' => $this->currency,
                        'value' => number_format($arBasket['arCalculation']['sum']['topay_brutto'], 2, '.', '')
                    ]
                ]],
                "application_context" => $application_context
            ];

            $subTotal = 0; $tax_total = 0; $item_total = 0; $discounts = 0;
            $items = []; $breakdown = [];

            foreach ($arBasket['produkte'] as $k => $p) {
 
				// Hier muss gerundet weiter gerechnet werden da die Summe am Ende von PayPal kontrolliert wird
				$p['preis_netto'] = round($p['preis_netto'], 2);
				$p['preis_brutto'] = round($p['preis_brutto'], 2);
				
                $sku = $this->shop->getProductAnr($this->shop->getProduktID($p['id']));
                if (!wpsg_isSizedString($sku)) $sku = __('n.A.', 'wpsg');

                $tax = number_format($p['preis_brutto'] - $p['preis_netto'], 2, '.', '');

                $items[] = [
                    'quantity' => $p['menge'],
                    'name' => $this->shop->getProductName($this->shop->getProduktID($p['id']), true),
                    'description' => '',
                    'sku' => $sku,
                    'unit_amount' => [
                        'currency_code' => $this->currency,
                        'value' => number_format($p['preis_netto'], 2, '.', '')
                    ],
                    'tax' => [
                        'currency_code' => $this->currency,
                        'value' => $tax
                    ]
                ];

                $item_total += $p['preis_netto'] * $p['menge'];
                $subTotal += $p['preis_brutto'] * $p['menge'];
                $tax_total += $tax * $p['menge'];

            }
		
		// PayPal kann noch keine Negativen "Items" anlegen
		// Must be greater than or equal to 0. If the currency supports decimals, only two decimal place precision is supported.
		/*
			if (isset($arBasket['arCalculation']['coupon'])) {
				foreach ($arBasket['arCalculation']['coupon'] as $k => $c) {

					$items[] = [
						'quantity' => '1',
						'name' => 'Wertgutschein',
						'description' => $c['code'],
						'sku' => $sku,
						'unit_amount' => [
                        	'currency_code' => $this->currency,
                        	'value' => number_format($c['netto'], 2, '.', '')
						],
						'tax' => [
                        	'currency_code' => $this->currency,
                        	'value' => $c['tax']
                    	]
						];

					$item_total += -$c['netto'];
					$subTotal += -$c['brutto'];
					$tax_total += -$c['tax'];

				} 
			}
		*/

            if (isset($arBasket['arCalculation']['shipping']['brutto']) && $arBasket['arCalculation']['shipping']['brutto'] > 0) {

                $items[] = [
                    'quantity' => 1,
                    'name' => __('Versandkosten', 'wpsg'),
                    'description' => '',
                    'sku' => '',
                    'unit_amount' => [
                        'currency_code' => $this->currency,
                        'value' => number_format($arBasket['arCalculation']['shipping']['netto'], 2, '.', '')
                    ],
                    'tax' => [
                        'currency_code' => $this->currency,
                        'value' => number_format($arBasket['arCalculation']['shipping']['brutto'] - $arBasket['arCalculation']['shipping']['netto'], 2, '.', '')
                    ]
                ];

                $item_total += $arBasket['arCalculation']['shipping']['netto'];
                $subTotal += $arBasket['arCalculation']['shipping']['brutto'];
                $tax_total += ($arBasket['arCalculation']['shipping']['brutto'] - $arBasket['arCalculation']['shipping']['netto']);

            }

            if (isset($arBasket['arCalculation']['sum']['payment_brutto']) && $arBasket['arCalculation']['sum']['payment_brutto'] > 0) {

                $items[] = [
                    'quantity' => 1,
                    'name' => __('Kosten für Zahlungsart', 'wpsg'),
                    'description' => '',
                    'sku' => '',
                    'unit_amount' => [
                        'currency_code' => $this->currency,
                        'value' => number_format($arBasket['arCalculation']['sum']['payment_netto'], 2, '.', '')
                    ],
                    'tax' => [
                        'currency_code' => $this->currency,
                        'value' => number_format($arBasket['arCalculation']['sum']['payment_brutto'] - $arBasket['arCalculation']['sum']['payment_netto'], 2, '.', '')
                    ]
                ];

                $item_total += $arBasket['arCalculation']['sum']['payment_netto'];
                $subTotal += $arBasket['arCalculation']['sum']['payment_brutto'];
                $tax_total += ($arBasket['arCalculation']['sum']['payment_brutto'] - $arBasket['arCalculation']['sum']['payment_netto']);

            }

			if (isset($arBasket['arCalculation']['voucher'])) {

				foreach ($arBasket['arCalculation']['voucher'] as $v) {

					$subTotal += $v['brutto'];

					$discounts += abs($v['brutto']);

				}

			}
			
			if (isset($arBasket['arCalculation']['coupon'])) {

				foreach ($arBasket['arCalculation']['coupon'] as $v) {

					$subTotal += $v['brutto'];

					$discounts += abs($v['brutto']);

				}

			}
			

            if (isset($arBasket['arCalculation']['discount'])) {

                foreach ($arBasket['arCalculation']['discount'] as $d) {

                    $subTotal += $d['brutto'];
  
                    $discounts += abs($d['brutto']);

                }

            }

            // amount/value
            // Should equal item_total + tax_total + shipping + handling + insurance - shipping_discount - discount.

			//wpsg_debug($arBasket);
            //wpsg_debug($arBasket['arCalculation']['sum']);
        	//wpsg_debug($items);
			//wpsg_debug(floatval($subTotal) === floatval($arBasket['arCalculation']['sum']['brutto']));
			//die;

            if (wpsg_isSizedInt($this->shop->get_option('wpsg_mod_paypalapi_paypalexpress_details')) && floatval($subTotal) === floatval($arBasket['arCalculation']['sum']['brutto'])) {

                $request->body['purchase_units'][0]['items'] = $items;

                $request->body['purchase_units'][0]['amount']['breakdown'] = [
                    'item_total' => [
                        'currency_code' => $this->currency,
                        'value' => 0
                    ],
                    'tax_total' => [
                        'currency_code' => $this->currency,
                        'value' => 0
                    ],
                    'shipping' => [
                        'currency_code' => $this->currency,
                        'value' => 0
                    ],
                    'handling' => [
                        'currency_code' => $this->currency,
                        'value' => 0
                    ],
                    'insurance' => [
                        'currency_code' => $this->currency,
                        'value' => 0
                    ],
                    'shipping_discount' => [
                        'currency_code' => $this->currency,
                        'value' => 0
                    ],
                    'discount' => [
                        'currency_code' => $this->currency,
                        'value' => 0
                    ]
                ];

                $request->body['purchase_units'][0]['amount']['breakdown']['item_total']['value'] =
                    number_format($item_total, 2, '.', '');

                $request->body['purchase_units'][0]['amount']['breakdown']['tax_total']['value'] =
                    number_format($tax_total, 2, '.', '');

                $request->body['purchase_units'][0]['amount']['breakdown']['discount']['value'] =
                    number_format($discounts, 2, '.', '');			
					
            }

            //wpsg_debug($request->body);

            try {

                $response = $client->execute($request);
                $response = json_decode(json_encode($response), true)['result'];

                if (wpsg_isSizedString($response['id'])) $oOrder->setMeta('wpsg_mod_paypalapi_paymentid', $response['id']);

                //echo '<pre>'.print_r($response, true).'</pre>';

                return $response;

            } catch (Exception $ex) {

                wpsg_debug(json_decode($ex->getMessage(), true));

                throw new \wpsg\Exception($ex->getMessage());

            }

            die("-");

        }

		/**
         * DEPRECATED
		 * Informationen zu einer authorisierten Zahlung holen
		 * @param unknown $payment_id
		 */
		private function api_getPaymentInfo($payment_id, $noError = false) {

			$json_data = $this->api_call('v1/payments/payment/'.$payment_id, false, $this->api_getAccessToken(), false, $noError);
									
			return $json_data;
			
		} // private function api_getPaymentInfo($payment_id)

        /**
         * Informationen zu einer authorisierten Zahlung holen
         * @param unknown $payment_id
         * @throws \Exception
         */
        private function api_getPaymentInfo_v2($payment_id, $noError = false) {

            if ($this->shop->get_option('wpsg_mod_paypalapi_version') === 'v1') {

                return $this->api_getPaymentInfo($payment_id, $noError);

            }

            $client = new \PayPalCheckoutSdk\Core\PayPalHttpClient($this->api_getEnvironment());
            $json_data = $client->execute(new \PayPalCheckoutSdk\Orders\OrdersGetRequest($payment_id));
            $json_data = json_decode(json_encode($json_data), true)['result'];

            //echo '<pre>'; print_r($json_data); echo '</pre>';

            return $json_data;

        }

		/**
		 * Generiert einen Access Token
		 */
		private function api_getAccessToken()
		{
						
			$json_data = $this->api_call('v1/oauth2/token', http_build_query(array('grant_type' => 'client_credentials')));

			return $json_data['access_token'];
			
		} // private function api_getAccessToken()
		
		/**
		 * Zentrale Funktion für Anfragen an die API
		 */
		private function api_call($command, $post_data, $auth_token = false, $patch = false, $noError = false, $put = false)
		{
						
			$ch = curl_init();
			
			curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
			
			if ($this->isSandbox()) curl_setopt($ch, CURLOPT_URL, "https://api.sandbox.paypal.com/".$command);
			else curl_setopt($ch, CURLOPT_URL, "https://api.paypal.com/".$command);
			
			$arHeader = array(
				'Accept: application/json',
				'Accept-Language: de_DE',
                'Prefer: return=representation'
			);
			
			if ($auth_token === false) {
						
				if ($this->isSandbox()) {
					 
					if (!wpsg_isSizedString($this->shop->get_option('wpsg_mod_paypalapi_sandbox_clientid')) || !wpsg_isSizedString($this->shop->get_option('wpsg_mod_paypalapi_sandbox_secret'))) {
						
						$this->api_handleError(__('Keine API Daten hinterlegt.', 'wpsg'));
						return;
						
					}
					 
					curl_setopt($ch, CURLOPT_USERPWD, $this->shop->get_option('wpsg_mod_paypalapi_sandbox_clientid').':'.$this->shop->get_option('wpsg_mod_paypalapi_sandbox_secret'));
					
				} else {
					
					if (!wpsg_isSizedString($this->shop->get_option('wpsg_mod_paypalapi_clientid')) || !wpsg_isSizedString($this->shop->get_option('wpsg_mod_paypalapi_secret'))) {
					
						$this->api_handleError(__('Keine API Daten hinterlegt.', 'wpsg'));
						return;
						
					}
					
					curl_setopt($ch, CURLOPT_USERPWD, $this->shop->get_option('wpsg_mod_paypalapi_clientid').':'.$this->shop->get_option('wpsg_mod_paypalapi_secret'));
					
				}
				
			}
			else
			{
			 
				$arHeader[] = 'Authorization: Bearer '.$auth_token;
				
			}
			
			if ($patch === true)
			{
				
				curl_setopt($ch, CURLOPT_CUSTOMREQUEST, 'PATCH');
			
			} 
			else if ($put === true)
			{
					
				curl_setopt($ch, CURLOPT_CUSTOMREQUEST, 'PUT');
			
			}
			else if ($post_data !== false)
			{
				
				curl_setopt($ch, CURLOPT_POST, true);
				
			}
			
			if (is_array($post_data))
			{
			
				$arHeader[] = 'Content-Type: application/json';
				 
				$post_data = json_encode($post_data, JSON_UNESCAPED_SLASHES);
 			
			}
			
			if ($post_data !== false) curl_setopt($ch, CURLOPT_POSTFIELDS, $post_data);

			curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
			curl_setopt($ch, CURLOPT_HTTPHEADER, $arHeader);
            curl_setopt($ch, CURLOPT_HEADER, 0);
									
			$output = curl_exec($ch);
		 
			if ($auth_token === false && $output == '')
			{
				
				$this->api_handleError(__('API lieferte keinen AUTH Token. Vermutlich fehlerhafte Zugangsdaten.', 'wpsg'));
				return;
				
			}
			
			$curl_error = curl_error($ch);
	 
			if (wpsg_isSizedString($curl_error)) $this->api_handleError($curl_error);
			
			curl_close($ch);

			$return = json_decode($output, true);
			
			if ($noError === true) return $return;
			
			if (is_array($return) && wpsg_isSizedString($return['error_description'])) $this->api_handleError($return['error_description']);
			
			if (!is_array($return) || wpsg_isSizedString($return['debug_id'])) $this->api_handleError($return);
			
			return $return;
			
		} // private function api_call($command)
		
		/**
		 * Fehlerbehandlung bei API Fehlern
		 */
		private function api_handleError($errorValue)
		{ 

			if ($this->api_try === true) return;
						
			if (is_array($errorValue)) $errorValue = print_r($errorValue, 1);
			
			// wpsg_debug($errorValue);
			
			if (wpsg_isSizedInt($_SESSION['wpsg']['order_id']))
			{
					 
				$oOrder = wpsg_order::getInstance($_SESSION['wpsg']['order_id']);
				$oOrder->log('PayPal API Fehler', $errorValue);
				
			}
			else if (wpsg_isSizedInt($_REQUEST['edit_id']))
			{
				
				$oOrder = wpsg_order::getInstance(wpsg_sinput("key", $_REQUEST['edit_id']));
				$oOrder->log('PayPal API Fehler', $errorValue);
				
			}
			
			if (!is_admin()) die();
			
		} // private function api_handleError($errorValue)

        /**
         * Autoloader für die PayPal Klassen
         */
        function spl_autoload($class) {

            $path = null;

            if (substr($class, 0, 18) === "PayPalCheckoutSdk\\") {

                $path = WPSG_PATH_MOD.'mod_paypalapi'.DIRECTORY_SEPARATOR.str_replace("\\", DIRECTORY_SEPARATOR, $class).'.php';

            } else if (substr($class, 0, 11) === "PayPalHttp\\") {

                $path = WPSG_PATH_MOD.'mod_paypalapi'.DIRECTORY_SEPARATOR.str_replace("\\", DIRECTORY_SEPARATOR, $class).'.php';

            }

            if ($path !== null && file_exists($path)) require_once($path);

        } // function spl_autoload($class)

	} // class wpsg_mod_paypalapi extends wpsg_mod_basic
